import { debug } from './logger'

import { createJsonDatabase } from './lib/file-db'

import { randomUUID } from 'crypto'
import { createSimpleRunner, type Runner } from './runners'

import { createWriteStream } from 'fs'
import { readFile, mkdir } from 'fs/promises'

import { dirname } from 'path'

import _ from 'lodash'
import { OnJobAdded, OnJobCompleted, OnJobLog, OnJobStarted } from './events'

type JobStatus = 'queued' | 'running' | 'completed'

export type Job = {
    uuid: string
    name: string

    status: JobStatus

    submitter: any
    submittedAt: string

    successful?: boolean
    error?: any
    startedAt?: string
    completedAt?: string
}

export type DeployFunction = (runner: Runner) => Promise<void>

type QueuedJob = {
    uuid: string
    deployFn: DeployFunction
    completed?: () => void
}

const jobsDB = createJsonDatabase<Job[]>(`${import.meta.env.DATA_PATH}/jobs.json`, [])

export function getJobLogFile(job: Job): string {
    const timeHash = new Date(job.submittedAt).getTime()

    // TODO: job.name is not safe
    return `${import.meta.env.DATA_PATH}/logs/${job.name}_${timeHash}_${job.uuid.slice(0, 6)}.log`
}

// To ensure that the while loop inside "processQueue" is getting executed from
// only one call at a time.
let working = false
const queue: QueuedJob[] = []

async function processQueue() {
    if (working) return

    working = true
    {
        while (queue.length > 0) {
            const { uuid, deployFn, completed } = queue.shift()!

            const job = await jobsDB.update(async jobs => {
                const job = jobs.find(job => job.uuid === uuid)!
                job.status = 'running'

                return job
            })

            OnJobStarted.emit(job)

            const startedAt = new Date().toISOString()
            let error: string | undefined

            const logsFilename = getJobLogFile(job)
            await mkdir(dirname(logsFilename), { recursive: true })

            const logsFile = createWriteStream(logsFilename, { flags: 'w' })

            const runner = createSimpleRunner(line => {
                logsFile.write(line + '\n')

                // to not block log file creation
                setImmediate(() => OnJobLog.emit(uuid, line))
            })

            debug(`[Jobs] Starting job "${job.name}"`)
            try {
                await deployFn(runner)
            } catch (e) {
                error = e!.toString()
            }

            if (error) {
                debug(`[Jobs] Error in job "${job.name}"`)
                debug(`[Jobs] ${error}`)
            }

            debug(`[Jobs] Finished job`)

            const completedAt = new Date().toISOString()

            const completedJob = await jobsDB.update(async jobs => {
                const job = jobs.find(job => job.uuid === uuid)!

                job.status = 'completed'

                job.successful = error === undefined
                job.error = error

                job.startedAt = startedAt
                job.completedAt = completedAt

                return job
            })

            completed?.()
            OnJobCompleted.emit(completedJob)
        }
    }
    working = false
}

export type JobBase = {
    name: string
    submitter: any
}

// Use this function to add new jobs to the work queue
export async function enqueueJob(
    jobBase: JobBase,
    deployFn: DeployFunction,
    completed?: () => void
) {
    const uuid = randomUUID()
    const job: Job = {
        ...jobBase,
        uuid,
        status: 'queued',
        submittedAt: new Date().toISOString(),
    }

    await jobsDB.update(async jobs => {
        jobs.push(job)
    })

    queue.push({ uuid, deployFn, completed })
    OnJobAdded.emit(job)

    //  starts concurrently a function to process jobs
    processQueue()
}

export function getJobs(): Promise<Job[]> {
    return jobsDB.load()
}

export async function getJob(uuid: string): Promise<Job> {
    const jobs = await jobsDB.load()
    const job = jobs.find(job => job.uuid === uuid)
    if (!job) throw new Error(`no job with uuid "${uuid}"`)
    return job
}

export async function getJobLogs(uuid: string): Promise<string> {
    const job = await getJob(uuid)

    try {
        return await readFile(getJobLogFile(job), 'utf8')
    } catch {
        return ''
    }
}

export async function getLatestJobByName(name: string): Promise<Job | undefined> {
    const jobs = await jobsDB.load()
    const jobsByName = jobs.filter(job => job.name === name)
    const latestJobs = _.sortBy(jobsByName, job => new Date(job.submittedAt))
    return latestJobs.length > 0 ? latestJobs.at(-1) : undefined
}

// ===================[ Old Version ]===================
/*

export type QueuedJob = {
    uuid: string
    name: string

    status: JobStatus

    submitter: any
    submittedAt: string
}

export type CompletedJob = {
    uuid: string
    name: string

    status: JobStatus

    successful: boolean
    error?: any

    submitter: any
    submittedAt: string
    startedAt: string
    completedAt: string
}

export type Job = {
    name: string

    status: JobStatus

    submitter: any
    submittedAt: string
}

export type SubmittedJob = {
    uuid: string
    name: string

    status: JobStatus

    submitter: any
    submittedAt: string

    successful?: boolean
    error?: any
    startedAt?: string
    completedAt?: string
}

type JobBare = {
    name: string
    uuid: string

    status: JobStatus
}

export function getJobLogFile(job: JobBare): string {
    // TODO: job.name is not safe
    return `${import.meta.env.DATA_PATH}/logs/${job.name}.${job.uuid}.log`
}

export type Worker = {
    work: (runner: Runner) => Promise<void>
}

// Event emitter & Queue (runtime only)
const emitter = new EventEmitter<{
    'job:add': [QueuedJob]
    'job:completed': [CompletedJob]
    'job:log': [string]
}>()

const queue: (QueuedJob & Worker)[] = []

const runningJobs: Record<string, SubmittedJob> = {}

// Job db for logging purposes
const jobsDB = createJsonDatabase<CompletedJob[]>(`${import.meta.env.DATA_PATH}/jobs.json`, [])

// To ensure that the while loop inside "processQueue" is getting executed from
// only one call at a time.
let working = false

async function processQueue() {
    if (working) return

    working = true
    {
        while (queue.length > 0) {
            const job = queue.shift()!

            const startedAt = new Date().toISOString()
            let error: string | undefined

            const logsFilename = getJobLogFile(job)

            await mkdir(dirname(logsFilename), { recursive: true })

            const logsFile = createWriteStream(logsFilename, { flags: 'w' })

            const runner = createSimpleRunner(line => {
                logsFile.write(line + '\n')

                // to not block log file creation
                setImmediate(() => emitter.emit('job:log', line))
            })

            debug(`[Jobs] Starting job "${job.name}"`)
            try {
                await job.work(runner)
            } catch (e) {
                error = e!.toString()
            }
            debug(`[Jobs] Finished job`)

            const completedAt = new Date().toISOString()

            const completedJob: CompletedJob = {
                uuid: job.uuid,
                name: job.name,

                status: 'completed',

                successful: error === undefined,
                error,

                submitter: job.submitter,
                submittedAt: job.submittedAt,
                startedAt,
                completedAt,
            }

            await jobsDB.update(async jobs => {
                jobs.push(completedJob)
            })

            emitter.emit('job:completed', completedJob)
        }
    }
    working = false
}

// Use this function to add new jobs to the work queue
export function enqueueJob(job: Job & Worker) {
    const queueJob = { ...job, uuid: randomUUID() }

    queue.push(queueJob)

    emitter.emit('job:add', {
        uuid: queueJob.uuid,
        name: queueJob.name,
        status: 'queued',
        submitter: queueJob.submitter,
        submittedAt: queueJob.submittedAt,
    })

    //  starts concurrently a function to process jobs
    processQueue()
}

export async function getQueuedJobs(): Promise<QueuedJob[]> {
    return queue.map(({ uuid, name, status, submitter, submittedAt }) => ({
        uuid,
        name,
        status,
        submitter,
        submittedAt,
    }))
}

export function getCompletedJobs(): Promise<CompletedJob[]> {
    return jobsDB.load()
}

export async function getJob(uuid: string): Promise<SubmittedJob | undefined> {
    const jobs = await jobsDB.load()
    const job = jobs.find(job => job.uuid === uuid)

    if (job) return job

    const queuedJob = queue.find(job => job.uuid === uuid)
    return queuedJob
}

export async function getJobLogs(job: JobBare): Promise<string> {
    return await readFile(getJobLogFile(job), 'utf8')
}

export const OnJobAdded = {
    addListener(cb: (job: QueuedJob) => void) {
        emitter.on('job:add', cb)
    },
    removeListener(cb: (job: QueuedJob) => void) {
        emitter.off('job:add', cb)
    },
}

export const OnJobCompleted = {
    addListener(cb: (job: CompletedJob) => void) {
        emitter.on('job:completed', cb)
    },
    removeListener(cb: (job: CompletedJob) => void) {
        emitter.off('job:completed', cb)
    },
}

export const OnJobLog = {
    addListener(cb: (line: string) => void) {
        emitter.on('job:log', cb)
    },
    removeListener(cb: (line: string) => void) {
        emitter.off('job:log', cb)
    },
}

*/
